from .bace import load_bace
from .bbbp import load_bbbp
from .benchmark import load_moleculenet_benchmark, load_ogb_splits
from .clintox import load_clintox
from .esol import load_esol
from .freesolv import load_freesolv
from .hiv import load_hiv
from .lipophilicity import load_lipophilicity
from .muv import load_muv
from .pcba import load_pcba
from .sider import load_sider
from .tox21 import load_tox21
from .toxcast import load_toxcast
